import codecs
import collections
import io
import os
import os.path
import shutil
import string
import subprocess
import sys
import yaml
import json
import signal
from datetime import datetime
from threading import Thread

from honcho import procfile
from tsuru_unit_agent.stream import Stream

WATCHER_TEMPLATE = u"""
[watcher:{name}]
cmd = {cmd}
copy_env = True
uid = {user}
gid = {group}
working_dir = {working_dir}
stdout_stream.class = tsuru.stream.Stream
stdout_stream.watcher_name = {name}
stderr_stream.class = tsuru.stream.Stream
stderr_stream.watcher_name = {name}
"""


def process_output(in_fd, out_fd):
    for line in iter(in_fd.readline, b''):
        if line is None:
            break
        out_fd.write(line)
    out_fd.flush()
    in_fd.close()
    out_fd.close()

running_pipe = None


def sigterm_handler(signum, sigframe):
    try:
        running_pipe.send_signal(signal.SIGTERM)
    except:
        pass
    sys.exit(0)
signal.signal(signal.SIGTERM, sigterm_handler)


def exec_with_envs(commands, with_shell=False, working_dir="/home/application/current", pipe_output=False,
                   envs=None):
    global running_pipe
    if not envs:
        envs = {}
    app_envs = {}
    app_envs.update(os.environ)
    app_envs.update(envs)
    if not os.path.exists(working_dir):
        working_dir = "/"
    for command in commands:
        popen_output = None
        if pipe_output:
            popen_output = subprocess.PIPE
        pipe = subprocess.Popen(command, shell=with_shell, cwd=working_dir, env=app_envs,
                                stdout=popen_output, stderr=popen_output)
        running_pipe = pipe
        if pipe_output:
            stdout = Stream(echo_output=sys.stdout,
                            default_stream_name='stdout',
                            watcher_name='unit-agent',
                            envs=app_envs)
            stderr = Stream(echo_output=sys.stderr,
                            default_stream_name='stderr',
                            watcher_name='unit-agent',
                            envs=app_envs)
            stdout_thread = Thread(target=process_output, args=(pipe.stdout, stdout))
            stdout_thread.start()
            stderr_thread = Thread(target=process_output, args=(pipe.stderr, stderr))
            stderr_thread.start()
        status = pipe.wait()
        running_pipe = None
        if pipe_output:
            stdout_thread.join()
            stderr_thread.join()
        if status != 0:
            sys.exit(status)


def execute_start_script(start_cmd, envs=None, with_shell=True):
    exec_with_envs([start_cmd], with_shell=with_shell, envs=envs)


def run_build_hooks(app_data, envs=None):
    commands = (app_data.get('hooks') or {}).get('build') or []
    exec_with_envs(commands, with_shell=True, envs=envs)


def run_restart_hooks(position, app_data, envs=None):
    restart_hook = (app_data.get('hooks') or {}).get('restart') or {}
    commands = restart_hook.get('{}-each'.format(position)) or []
    commands += restart_hook.get(position) or []
    exec_with_envs(commands, with_shell=True, pipe_output=True,
                   envs=envs)


def load_app_yaml(working_dir="/home/application/current"):
    files_name = ["tsuru.yaml", "tsuru.yml", "app.yaml", "app.yml"]
    for file_name in files_name:
        try:
            fullpath = os.path.join(working_dir, file_name)
            with codecs.open(fullpath, 'r', encoding='utf-8', errors='ignore') as f:
                return yaml.load(f.read()) or {}
        except (IOError, yaml.scanner.ScannerError):
            pass
    return {}


def load_procfile(working_dir="/home/application/current"):
    path = os.path.join(working_dir, "Procfile")
    with codecs.open(path, "r", encoding="utf-8", errors="ignore") as f:
        return f.read()


def write_circus_conf(procfile_path=None, conf_path="/etc/circus/circus.ini",
                      envs=None):
    if not envs:
        envs = {}
    expanding_envs = collections.defaultdict(str)
    expanding_envs.update(os.environ)
    expanding_envs.update(envs)
    procfile_path = procfile_path or os.environ.get("PROCFILE_PATH",
                                                    "/home/application/current/Procfile")
    content = ""
    with open(procfile_path) as f:
        content = f.read()
    pfile = procfile.Procfile(content)
    new_watchers = []
    working_dir = os.environ.get("APP_WORKING_DIR", "/home/application/current")
    for name, cmd in pfile.commands.items():
        cmd = string.Template(cmd).substitute(expanding_envs)
        new_watchers.append(WATCHER_TEMPLATE.format(name=name, cmd=cmd,
                                                    user="ubuntu", group="ubuntu",
                                                    working_dir=working_dir))
    base_conf = conf_path + ".base"
    if os.path.exists(conf_path):
        if not os.path.exists(base_conf):
            shutil.copy2(conf_path, base_conf)
        shutil.copy2(base_conf, conf_path)
    if new_watchers:
        with open(conf_path, "a") as f:
            for watcher in new_watchers:
                f.write(watcher)


def save_apprc_file(environs, file_path="/home/application/apprc"):
    with io.open(file_path, "w") as file:
        file.write(u"# generated by tsuru at {}\n".format(datetime.now()))
        for name, value in environs.iteritems():
            value = value.replace("'", "'\\''")
            file.write(u"export {}='{}'\n".format(name, value))


def parse_apprc_file(file_path="/home/application/apprc"):
    # Let the shell parse environment variables for us, there are escaping
    # edge cases we don't want to be aware of.
    pipe = subprocess.Popen('source {}; python -c "import os, json; e=dict(); '.format(file_path) +
                            'e.update(os.environ); print json.dumps(e)"',
                            shell=True, env={}, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    result = pipe.communicate()
    return json.loads(result[0])



# -*- coding: utf-8 -*-

from unittest import TestCase
import mock
import os
import sys

from tsuru_unit_agent.tasks import (
    execute_start_script,
    load_app_yaml,
    load_procfile,
    run_build_hooks,
    run_restart_hooks,
    save_apprc_file,
    parse_apprc_file,
    write_circus_conf,
)


class TestTasks(TestCase):

    @mock.patch("os.environ", {'env': 'var', 'env1': 'var1'})
    @mock.patch("subprocess.Popen")
    def test_execute(self, popen_mock):
        wait_mock = popen_mock.return_value.wait
        wait_mock.return_value = 0
        execute_start_script("my_command", envs={"env": "varrr"})
        self.assertEqual(popen_mock.call_args[0][0], 'my_command')
        self.assertEqual(popen_mock.call_args[1]['shell'], True)
        self.assertEqual(popen_mock.call_args[1]['cwd'], '/')
        self.assertDictEqual(popen_mock.call_args[1]['env'], {'env': 'varrr',
                                                              'env1': 'var1'})
        wait_mock.assert_called_once_with()

    @mock.patch("os.environ", {'myenv': 'var', 'env1': 'var1'})
    @mock.patch("sys.exit")
    @mock.patch("subprocess.Popen")
    def test_execute_failing(self, popen_mock, exit_mock):
        wait_mock = popen_mock.return_value.wait
        wait_mock.return_value = 10
        execute_start_script("my_command", envs={"myenv": "varrr"})
        self.assertEqual(popen_mock.call_args[0][0], 'my_command')
        self.assertEqual(popen_mock.call_args[1]['shell'], True)
        self.assertEqual(popen_mock.call_args[1]['cwd'], '/')
        self.assertDictEqual(popen_mock.call_args[1]['env'], {'myenv': 'varrr',
                                                              'env1': 'var1'})
        wait_mock.assert_called_once_with()
        exit_mock.assert_called_once_with(10)

    def test_save_apprc_file(self):
        environs = {"DATABASE_HOST": "localhost", "DATABASE_USER": "root"}
        with mock.patch("io.open", mock.mock_open()) as m:
            save_apprc_file(environs)
            m.assert_called_once_with("/home/application/apprc", "w")
            write_mock = m().write
            self.assertRegexpMatches(write_mock.mock_calls[0][1][0],
                                     '# generated by tsuru at .*\n')
            write_mock.assert_any_call('export DATABASE_HOST=\'localhost\'\n')
            write_mock.assert_any_call('export DATABASE_USER=\'root\'\n')
            self.assertEqual(len(write_mock.mock_calls), 3)

    def test_parse_apprc_file(self):
        path = os.path.join(os.path.dirname(__file__), "fixtures", "apprc")
        envs = parse_apprc_file(path)
        expected = {
            "A": "B",
            "C": "C D",
            "b": "888",
            "B": "9\"1",
            "D": "X=y",
            "F": "a(a",
            "MY_awesome_BIG_name": "something",
        }
        for k, v in envs.items():
            if k not in expected:
                envs.pop(k)
        self.assertDictEqual(envs, expected)

    def test_save_parse_apprc_escaping(self):
        expected = {
            "A": "B",
            "C": "C D",
            "b": "888",
            "B": "9\"1",
            "D": "X=y",
            "F": "a(a) + (",
            "MY_awesome_BIG_name": "some'thin'g'",
            "JSON": '[{"a": "b", "c": {"d": "f"}}]',
            "SLASHED": "a\\a escaped \\\" some \\\"",
            "EXEC": "a: `echo hey` b: $(echo again)",
            "MULTILINE": "my\nmulti\"line\", with ' quotes ' yay'\nvariable'\n'",
        }
        path = os.path.join(os.path.dirname(__file__), "fixtures", "written_apprc")
        try:
            save_apprc_file(expected, file_path=path)
            envs = parse_apprc_file(path)
            for k, v in envs.items():
                if k not in expected:
                    envs.pop(k)
            self.assertDictEqual(envs, expected)
        finally:
            os.remove(path)


class RunHooksTest(TestCase):

    @mock.patch("os.environ", {'env': 'var', 'env1': 'var1'})
    @mock.patch("subprocess.Popen")
    def test_execute_commands(self, popen_call):
        wait_mock = popen_call.return_value.wait
        wait_mock.return_value = 0
        data = {"hooks": {"build": ["ble"]}}
        run_build_hooks(data, envs={'env': 'varrr'})
        self.assertEqual(popen_call.call_args[0][0], 'ble')
        self.assertEqual(popen_call.call_args[1]['shell'], True)
        self.assertEqual(popen_call.call_args[1]['cwd'], '/')
        self.assertDictEqual(popen_call.call_args[1]['env'], {'env': 'varrr',
                                                              'env1': 'var1'})
        wait_mock.assert_called_once_with()

    @mock.patch("os.environ", {})
    @mock.patch("os.path.exists")
    @mock.patch("subprocess.Popen")
    def test_execute_commands_default_cwd_exists(self, popen_call, exists_mock):
        wait_mock = popen_call.return_value.wait
        wait_mock.return_value = 0
        exists_mock.return_value = True
        data = {"hooks": {"build": ["ble"]}}
        run_build_hooks(data)
        self.assertEqual(popen_call.call_args[0][0], 'ble')
        self.assertEqual(popen_call.call_args[1]['shell'], True)
        self.assertEqual(popen_call.call_args[1]['cwd'], '/home/application/current')
        self.assertDictEqual(popen_call.call_args[1]['env'], {})
        wait_mock.assert_called_once_with()
        exists_mock.assert_called_once_with("/home/application/current")

    @mock.patch("os.environ", {})
    @mock.patch("sys.exit")
    @mock.patch("subprocess.Popen")
    def test_execute_failing_commands(self, popen_call, exit_mock):
        wait_mock = popen_call.return_value.wait
        wait_mock.return_value = 5
        data = {"hooks": {"build": ["ble"]}}
        run_build_hooks(data)
        self.assertEqual(popen_call.call_args[0][0], 'ble')
        self.assertEqual(popen_call.call_args[1]['shell'], True)
        self.assertEqual(popen_call.call_args[1]['cwd'], '/')
        self.assertDictEqual(popen_call.call_args[1]['env'], {})
        wait_mock.assert_called_once_with()
        exit_mock.assert_called_once_with(5)

    @mock.patch("subprocess.Popen")
    def test_execute_commands_hooks_empty(self, subprocess_call):
        data = {}
        run_build_hooks(data)
        subprocess_call.assert_not_called()
        data = {"hooks": None}
        run_build_hooks(data)
        subprocess_call.assert_not_called()
        data = {"hooks": {"build": None}}
        run_build_hooks(data)
        subprocess_call.assert_not_called()
        data = {"hooks": {"build": []}}
        run_build_hooks(data)
        subprocess_call.assert_not_called()


class RunRestartHooksTest(TestCase):

    @mock.patch("tsuru_unit_agent.tasks.Stream")
    @mock.patch("os.environ", {'env': 'var', 'env1': 'var1'})
    @mock.patch("subprocess.Popen")
    def test_run_restart_hooks(self, popen_call, Stream):
        popen_call.return_value.stdout.readline.return_value = ''
        popen_call.return_value.stderr.readline.return_value = ''
        wait_mock = popen_call.return_value.wait
        wait_mock.return_value = 0
        data = {"hooks": {"restart": {
            "before": ["b1"],
            "before-each": ["b2"],
            "after": ["a1"],
            "after-each": ["a2"],
        }}}
        run_restart_hooks('before', data, envs={'env': 'varrr'})
        self.assertEqual(popen_call.call_count, 2)
        self.assertEqual(popen_call.call_args_list[0][0][0], 'b2')
        self.assertEqual(popen_call.call_args_list[0][1]['shell'], True)
        self.assertEqual(popen_call.call_args_list[0][1]['cwd'], '/')
        self.assertDictEqual(popen_call.call_args_list[0][1]['env'], {'env': 'varrr',
                                                                      'env1': 'var1'})
        self.assertEqual(popen_call.call_args_list[1][0][0], 'b1')
        wait_mock.assert_any_call()
        run_restart_hooks('after', data)
        self.assertEqual(popen_call.call_count, 4)
        self.assertEqual(popen_call.call_args_list[3][0][0], 'a1')
        self.assertEqual(popen_call.call_args_list[2][0][0], 'a2')

    @mock.patch("tsuru_unit_agent.tasks.Stream")
    @mock.patch("os.environ", {'env': 'var', 'env1': 'var1'})
    @mock.patch("subprocess.Popen")
    def test_run_restart_hooks_calls_stream(self, popen_call, stream_mock):
        out1 = ['stdout_out1', '']
        out2 = ['stderr_out1', '']
        popen_call.return_value.stdout.readline.side_effect = lambda: out1.pop(0)
        popen_call.return_value.stderr.readline.side_effect = lambda: out2.pop(0)
        wait_mock = popen_call.return_value.wait
        wait_mock.return_value = 0
        data = {"hooks": {"restart": {
            "before": ["b1"],
        }}}
        run_restart_hooks('before', data, envs={'env': 'varrr'})
        self.assertDictEqual(os.environ, {'env': 'var', 'env1': 'var1'})
        self.assertEqual(popen_call.call_count, 1)
        self.assertEqual(popen_call.call_args_list[0][0][0], 'b1')
        self.assertEqual(popen_call.call_args_list[0][1]['shell'], True)
        self.assertEqual(popen_call.call_args_list[0][1]['cwd'], '/')
        self.assertDictEqual(popen_call.call_args_list[0][1]['env'], {'env': 'varrr',
                                                                      'env1': 'var1'})
        wait_mock.assert_called_once_with()
        self.assertEqual(stream_mock.call_count, 2)
        self.assertEqual(stream_mock.call_args_list[0][1]['echo_output'], sys.stdout)
        self.assertEqual(stream_mock.call_args_list[0][1]['default_stream_name'], 'stdout')
        self.assertEqual(stream_mock.call_args_list[0][1]['watcher_name'], 'unit-agent')
        self.assertEqual(stream_mock.call_args_list[0][1]['envs'], {'env': 'varrr',
                                                                    'env1': 'var1'})
        self.assertEqual(stream_mock.call_args_list[1][1]['echo_output'], sys.stderr)
        self.assertEqual(stream_mock.call_args_list[1][1]['default_stream_name'], 'stderr')
        self.assertEqual(stream_mock.call_args_list[1][1]['watcher_name'], 'unit-agent')
        self.assertEqual(stream_mock.call_args_list[1][1]['envs'], {'env': 'varrr',
                                                                    'env1': 'var1'})
        write_mock = stream_mock.return_value.write
        write_mock.assert_any_call('stdout_out1')
        write_mock.assert_any_call('stderr_out1')
        stream_mock.return_value.flush.assert_any_call()
        stream_mock.return_value.close.assert_any_call()


class LoadAppYamlTest(TestCase):

    def setUp(self):
        self.working_dir = os.path.dirname(__file__)
        self.data = '''
hooks:
  build:
    - {0}_1
    - {0}_2'''

    def test_load_app_yaml(self):
        filenames = ["tsuru.yaml", "tsuru.yml", "app.yaml", "app.yml"]
        for name in filenames:
            with open(os.path.join(self.working_dir, name), "w") as f:
                f.write(self.data.format(name))

        for name in filenames:
            data = load_app_yaml(self.working_dir)
            self.assertEqual(data, {"hooks": {"build": ["{}_1".format(name),
                                                        "{}_2".format(name)]}})
            os.remove(os.path.join(self.working_dir, name))

    def test_load_without_app_files(self):
        data = load_app_yaml(self.working_dir)
        self.assertDictEqual(data, {})

    def test_load_with_empty_yaml(self):
        with open(os.path.join(self.working_dir, "tsuru.yaml"), "w") as f:
            f.write("")
        data = load_app_yaml(self.working_dir)
        self.assertDictEqual(data, {})
        os.remove(os.path.join(self.working_dir, "tsuru.yaml"))

    def test_load_yaml_encoding(self):
        data = load_app_yaml(os.path.join(self.working_dir, "fixtures/iso88591"))
        self.assertDictEqual(data, {"key": "x"})
        data = load_app_yaml(os.path.join(self.working_dir, "fixtures/utf-8"))
        self.assertDictEqual(data, {"key": u"áéíãôüx"})

    def test_load_broken_yaml(self):
        broken_yaml = '''
hooks:
\tbuild:
\t\t- foo_1
\t\t- bar_2
        '''
        with open(os.path.join(self.working_dir, "tsuru.yaml"), "w") as f:
            f.write(broken_yaml)
        data = load_app_yaml(self.working_dir)
        self.assertDictEqual(data, {})
        os.remove(os.path.join(self.working_dir, "tsuru.yaml"))


class LoadProcfileTest(TestCase):

    def setUp(self):
        self.working_dir = os.path.join(os.path.dirname(__file__),
                                        "fixtures")

    def test_load_procfile(self):
        content = load_procfile(self.working_dir)
        expected = r"""web: python run_my_app.py -p $PORT -l $POORT
worker: python run_my_worker.py""" + "\n"
        self.assertEqual(expected, content)


class WriteCircusConfTest(TestCase):

    def setUp(self):
        self.procfile_path = os.path.join(os.path.dirname(__file__), "fixtures",
                                          "Procfile")
        self.conf_path = os.path.join(os.path.dirname(__file__), "fixtures",
                                      "circus.ini")
        self.original_conf = open(self.conf_path).read()
        os.environ["PORT"] = "9090"
        os.environ["POORT"] = "8989"

    def tearDown(self):
        open(self.conf_path, "w").write(self.original_conf)
        del os.environ["PORT"]
        del os.environ["POORT"]

    def test_write_file(self):
        expected_file = open(self.conf_path).read()
        expected_file += u"""
[watcher:web]
cmd = python run_my_app.py -p 8888 -l 8989
copy_env = True
uid = ubuntu
gid = ubuntu
working_dir = /home/application/current
stdout_stream.class = tsuru.stream.Stream
stdout_stream.watcher_name = web
stderr_stream.class = tsuru.stream.Stream
stderr_stream.watcher_name = web

[watcher:worker]
cmd = python run_my_worker.py
copy_env = True
uid = ubuntu
gid = ubuntu
working_dir = /home/application/current
stdout_stream.class = tsuru.stream.Stream
stdout_stream.watcher_name = worker
stderr_stream.class = tsuru.stream.Stream
stderr_stream.watcher_name = worker
"""
        write_circus_conf(procfile_path=self.procfile_path, conf_path=self.conf_path,
                          envs={"PORT": "8888"})
        got_file = open(self.conf_path).read()
        self.assertEqual(expected_file, got_file)

    def test_write_file_multiple_times(self):
        expected_file = open(self.conf_path).read()
        expected_file += u"""
[watcher:web]
cmd = python run_my_app.py -p 8888 -l 8989
copy_env = True
uid = ubuntu
gid = ubuntu
working_dir = /home/application/current
stdout_stream.class = tsuru.stream.Stream
stdout_stream.watcher_name = web
stderr_stream.class = tsuru.stream.Stream
stderr_stream.watcher_name = web

[watcher:worker]
cmd = python run_my_worker.py
copy_env = True
uid = ubuntu
gid = ubuntu
working_dir = /home/application/current
stdout_stream.class = tsuru.stream.Stream
stdout_stream.watcher_name = worker
stderr_stream.class = tsuru.stream.Stream
stderr_stream.watcher_name = worker
"""
        write_circus_conf(procfile_path=self.procfile_path, conf_path=self.conf_path,
                          envs={"PORT": "8888"})
        write_circus_conf(procfile_path=self.procfile_path, conf_path=self.conf_path,
                          envs={"PORT": "8888"})
        write_circus_conf(procfile_path=self.procfile_path, conf_path=self.conf_path,
                          envs={"PORT": "8888"})
        got_file = open(self.conf_path).read()
        self.assertEqual(expected_file, got_file)

    def test_write_file_no_watchers(self):
        expected_file = open(self.conf_path).read()
        write_circus_conf(procfile_path=self.procfile_path + ".empty",
                          conf_path=self.conf_path)
        got_file = open(self.conf_path).read()
        self.assertEqual(expected_file, got_file)

    def test_write_new_file(self):
        expected_file = u"""
[watcher:web]
cmd = python run_my_app.py -p 8888 -l 8989
copy_env = True
uid = ubuntu
gid = ubuntu
working_dir = /home/application/current
stdout_stream.class = tsuru.stream.Stream
stdout_stream.watcher_name = web
stderr_stream.class = tsuru.stream.Stream
stderr_stream.watcher_name = web

[watcher:worker]
cmd = python run_my_worker.py
copy_env = True
uid = ubuntu
gid = ubuntu
working_dir = /home/application/current
stdout_stream.class = tsuru.stream.Stream
stdout_stream.watcher_name = worker
stderr_stream.class = tsuru.stream.Stream
stderr_stream.watcher_name = worker
"""
        conf_path = self.conf_path + ".new"
        write_circus_conf(procfile_path=self.procfile_path,
                          conf_path=conf_path, envs={"PORT": "8888"})
        self.addCleanup(os.remove, conf_path)
        got_file = open(conf_path).read()
        self.assertEqual(expected_file, got_file)

    def test_write_new_file_env_working_dir(self):
        os.environ["APP_WORKING_DIR"] = "/tmp/app"

        def clean():
            del os.environ["APP_WORKING_DIR"]
        self.addCleanup(clean)
        expected_file = u"""
[watcher:web]
cmd = python run_my_app.py -p 8888 -l 8989
copy_env = True
uid = ubuntu
gid = ubuntu
working_dir = /tmp/app
stdout_stream.class = tsuru.stream.Stream
stdout_stream.watcher_name = web
stderr_stream.class = tsuru.stream.Stream
stderr_stream.watcher_name = web

[watcher:worker]
cmd = python run_my_worker.py
copy_env = True
uid = ubuntu
gid = ubuntu
working_dir = /tmp/app
stdout_stream.class = tsuru.stream.Stream
stdout_stream.watcher_name = worker
stderr_stream.class = tsuru.stream.Stream
stderr_stream.watcher_name = worker
"""
        conf_path = self.conf_path + ".new"
        write_circus_conf(procfile_path=self.procfile_path,
                          conf_path=conf_path, envs={"PORT": "8888"})
        self.addCleanup(os.remove, conf_path)
        got_file = open(conf_path).read()
        self.assertEqual(expected_file, got_file)

    def test_write_new_file_env_procfile_path(self):
        os.environ["PROCFILE_PATH"] = self.procfile_path + "2"

        def clean():
            del os.environ["PROCFILE_PATH"]
        expected_file = u"""
[watcher:web]
cmd = python run_their_app.py -p 8888
copy_env = True
uid = ubuntu
gid = ubuntu
working_dir = /home/application/current
stdout_stream.class = tsuru.stream.Stream
stdout_stream.watcher_name = web
stderr_stream.class = tsuru.stream.Stream
stderr_stream.watcher_name = web

[watcher:worker]
cmd = python run_their_worker.py
copy_env = True
uid = ubuntu
gid = ubuntu
working_dir = /home/application/current
stdout_stream.class = tsuru.stream.Stream
stdout_stream.watcher_name = worker
stderr_stream.class = tsuru.stream.Stream
stderr_stream.watcher_name = worker
"""
        conf_path = self.conf_path + ".new"
        write_circus_conf(conf_path=conf_path, envs={"PORT": "8888"})
        self.addCleanup(os.remove, conf_path)
        got_file = open(conf_path).read()
        self.assertEqual(expected_file, got_file)



# Copyright 2015 tsuru-unit-agent authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import unittest
import mock
import logging
import socket

from tsuru_unit_agent.stream import Stream, QUEUE_DONE_MESSAGE

mocked_environ = {
    "TSURU_APPNAME": "appname1",
    "TSURU_HOST": "host1",
    "TSURU_APP_TOKEN": "secret123",
    "TSURU_SYSLOG_SERVER": "host2",
    "TSURU_SYSLOG_PORT": "514",
    "TSURU_SYSLOG_FACILITY": "LOCAL0",
    "TSURU_SYSLOG_SOCKET": "udp",
}


@mock.patch("os.environ", mocked_environ)
class StreamTestCase(unittest.TestCase):

    @mock.patch("os.environ", mocked_environ)
    @mock.patch("tsuru_unit_agent.stream.gethostname")
    @mock.patch("tsuru_unit_agent.stream.TsuruLogWriter")
    def setUp(self, TsuruLogWriter, gethostname):
        TsuruLogWriter.return_value = log_writer = mock.Mock()
        gethostname.return_value = "myhost"
        l_out = "2012-11-06 17:13:55 [12019] [INFO] Starting gunicorn 0.15.0\n"
        l_err = "2012-11-06 17:13:55 [12019] [ERROR] Error starting gunicorn\n"
        self.data = {}
        self.data["stderr"] = {
            "pid": 12018,
            "data": l_err,
            "name": "stderr"
        }
        self.data["stdout"] = {
            "pid": 12018,
            "data": l_out,
            "name": "stdout"
        }
        self.stream = Stream(watcher_name="mywatcher")
        TsuruLogWriter.assert_called_with(mock.ANY, self.stream.queue)
        log_writer.start.assert_called_once()

    def tearDown(self):
        self.stream.writer.stop()

    def test_should_have_the_close_method(self):
        self.assertTrue(hasattr(Stream, "close"))

    def test_should_send_done_message_on_close(self):
        self.stream.close()
        entry = self.stream.queue.get()
        self.assertEqual(entry, QUEUE_DONE_MESSAGE)

    def test_should_send_log_to_tsuru(self):
        self.stream(self.data["stdout"])
        (appname, host, token, syslog_server,
         syslog_port, syslog_facility, syslog_socket) = self.stream._load_envs()
        url = "{0}/apps/{1}/log?source=mywatcher&unit=myhost".format(host,
                                                                     appname)
        expected_msg = "Starting gunicorn 0.15.0\n"
        entry = self.stream.queue.get()
        self.assertEqual(url, entry.url)
        self.assertEqual(2, entry.timeout)
        self.assertEqual([expected_msg], entry.messages)

    @mock.patch("logging.getLogger")
    @mock.patch("logging.handlers.SysLogHandler")
    def test_should_send_log_to_syslog_as_info(self, s_handler, logger):
        self.stream(self.data["stdout"])
        (appname, host, token, syslog_server,
         syslog_port, syslog_facility, syslog_socket) = self.stream._load_envs()
        my_logger = logger(appname)
        log_handler = s_handler(address=(syslog_server, syslog_port),
                                facility=syslog_facility,
                                socktype=syslog_socket)
        expected_msg = "Starting gunicorn 0.15.0\n"
        my_logger.addHandler(log_handler)
        my_logger.info.assert_called_with(expected_msg)

    @mock.patch("logging.getLogger")
    @mock.patch("tsuru_unit_agent.syslog.SysLogHandler")
    def test_should_send_log_to_syslog_as_error(self, s_handler, logger):
        s_handler.return_value = syslog = mock.Mock()
        self.stream(self.data["stderr"])
        (appname, host, token, syslog_server,
         syslog_port, syslog_facility, syslog_socket) = self.stream._load_envs()
        my_logger = logger(appname)
        s_handler.assert_called_with(address=(syslog_server, int(syslog_port)),
                                     facility=syslog_facility,
                                     socktype=socket.SOCK_DGRAM)
        my_logger.addHandler.assert_called_with(syslog)
        my_logger.error.assert_called_with("Error starting gunicorn\n")

    def test_should_send_log_to_syslog_and_use_one_handler(self):
        self.stream(self.data["stderr"])
        self.stream(self.data["stderr"])
        self.stream(self.data["stderr"])
        self.stream(self.data["stderr"])
        (appname, host, token, syslog_server,
         syslog_port, syslog_facility, syslog_socket) = self.stream._load_envs()
        my_logger = logging.getLogger(appname)
        self.assertEqual(len(my_logger.handlers), 1)

    @mock.patch("tsuru_unit_agent.stream.gethostname")
    @mock.patch("tsuru_unit_agent.stream.TsuruLogWriter")
    def test_timeout_is_configurable(self, TsuruLogWriter, gethostname):
        TsuruLogWriter.return_value = mock.Mock()
        gethostname.return_value = "myhost"
        stream = Stream(watcher_name="watcher", timeout=10)
        stream(self.data["stdout"])
        (appname, host, token, syslog_server,
         syslog_port, syslog_facility, syslog_socket) = self.stream._load_envs()
        url = "{0}/apps/{1}/log?source=watcher&unit=myhost".format(host,
                                                                   appname)
        expected_msg = "Starting gunicorn 0.15.0\n"
        entry = stream.queue.get(timeout=2)
        self.assertEqual(url, entry.url)
        self.assertEqual(10, entry.timeout)
        self.assertEqual([expected_msg], entry.messages)

    @mock.patch("tsuru_unit_agent.stream.gethostname")
    @mock.patch("tsuru_unit_agent.stream.TsuruLogWriter")
    def test_envs_may_be_overriden(self, TsuruLogWriter, gethostname):
        TsuruLogWriter.return_value = mock.Mock()
        gethostname.return_value = "myhost"
        stream = Stream(watcher_name="watcher", envs={"TSURU_APPNAME": "mynewappname"})
        stream(self.data["stdout"])
        (appname, host, token, syslog_server,
         syslog_port, syslog_facility, syslog_socket) = stream._load_envs()
        self.assertEqual(appname, "mynewappname")
        self.assertEqual(host, "host1")
        self.assertEqual(token, "secret123")
        self.assertEqual(syslog_server, "host2")
        self.assertEqual(syslog_port, "514")
        self.assertEqual(syslog_facility, "LOCAL0")
        self.assertEqual(syslog_socket, "udp")

    @mock.patch("tsuru_unit_agent.stream.TsuruLogWriter")
    @mock.patch("os.environ", {})
    def test_should_slience_errors_when_envs_does_not_exist(self, TsuruLogWriter):
        try:
            stream = Stream()
            stream(self.data["stdout"])
        except Exception as e:
            msg = "Should not fail when envs does not exist. " \
                  "Exception: {}".format(e)
            self.fail(msg)

    def test_get_messagess_no_buffering(self):
        msg = "2012-11-06 18:30:10 [13887] [INFO] Listening at: " \
              "http://127.0.0.1:8000 (13887)\n2012-11-06 18:30:10 [13887] " \
              "[INFO] Using worker: sync\n2012-11-06 18:30:10 [13890] " \
              "[INFO] Booting worker with pid: 13890\n2012-11-06 18:30:10 " \
              "[13890] [ERROR] Exception in worker process:\nTraceback " \
              "(most recent call last):\n"
        expected = [
            "Listening at: http://127.0.0.1:8000 (13887)\n",
            "Using worker: sync\n",
            "Booting worker with pid: 13890\n",
            "Exception in worker process:\n",
            "Traceback (most recent call last):\n",
        ]
        messages = self.stream._get_messages(msg)
        self.assertEqual("", self.stream._buffer)
        self.assertEqual(expected, messages)

    def test_get_messagess_buffering(self):
        msg = "2012-11-06 18:30:10 [13887] [INFO] Listening at: " \
              "http://127.0.0.1:8000 (13887)\n2012-11-06 18:30:10 [13887] " \
              "[INFO] Using worker: sync\n2012-11-06 18:30:10 [13890] " \
              "[INFO] Booting worker with pid: 13890\n2012-11-06 18:30:10 " \
              "[13890] [ERROR] Exception in worker process:\nTraceback " \
              "(most recent call last):"
        expected = [
            "Listening at: http://127.0.0.1:8000 (13887)\n",
            "Using worker: sync\n",
            "Booting worker with pid: 13890\n",
            "Exception in worker process:\n",
        ]
        messages = self.stream._get_messages(msg)
        self.assertEqual("Traceback (most recent call last):", self.stream._buffer)
        self.assertEqual(expected, messages)

    def test_get_messagess_buffered(self):
        msg1 = "2012-11-06 18:30:10 [13887] [INFO] Listening at: " \
               "http://127.0.0.1:8000 (13887)\n2012-11-06 18:30:10 [13887] "
        msg2 = "[INFO] Using worker: sync\n2012-11-06 18:30:10 "
        msg3 = "[13890] [INFO] Booting worker with pid: 13890\n2012-11-06 " \
               "18:30:10 [13890] [ERROR] Exception in worker process:\n" \
               "Traceback (most recent call last):\n"
        expected = [
            "Listening at: http://127.0.0.1:8000 (13887)\n",
            "Using worker: sync\n",
            "Booting worker with pid: 13890\n",
            "Exception in worker process:\n",
            "Traceback (most recent call last):\n",
        ]
        self.assertEqual(expected[:1], self.stream._get_messages(msg1))
        self.assertEqual(expected[1:2], self.stream._get_messages(msg2))
        self.assertEqual(expected[2:], self.stream._get_messages(msg3))

    def test_get_messagess_full_buffer(self):
        self.stream._max_buffer_size = 20
        self.stream._buffer = 13 * "a"
        msg = "2012-11-06 18:30:10 [13887] [INFO] Listening at: " \
              "http://127.0.0.1:8000 (13887)"
        expected = [self.stream._buffer +
                    "Listening at: http://127.0.0.1:8000 (13887)"]
        self.assertEqual(expected, self.stream._get_messages(msg))

    def test_get_messages_buffered_multiple_times(self):
        msg1 = "2012-11-06 18:30:10 [13887] [INFO] Listening at: " \
               "http://127.0.0.1:8000 (13887)\n2012-11-06 18:30:10 [13887] "
        msg2 = "[INFO] Using worker: sync\n2012-11-06 18:30:10 "
        msg3 = "[13890] [INFO] Booting worker with pid: 13890\n2012-11-06 " \
               "18:30:10 [13890] [ERROR] Exception in worker process:\n" \
               "Traceback (most recent call last):"
        msg4 = "\n2012-11-06 18:30:10 [13887] [INFO] Booting another worker "\
               "with pid: 13891\n"
        expected = [
            "Listening at: http://127.0.0.1:8000 (13887)\n",
            "Using worker: sync\n",
            "Booting worker with pid: 13890\n",
            "Exception in worker process:\n",
            "Traceback (most recent call last):\n",
            "Booting another worker with pid: 13891\n",
        ]
        self.assertEqual(expected[:1], self.stream._get_messages(msg1))
        self.assertEqual(expected[1:2], self.stream._get_messages(msg2))
        self.assertEqual(expected[2:4], self.stream._get_messages(msg3))
        self.assertEqual(expected[4:], self.stream._get_messages(msg4))

    def test_get_messages_sequential_bufferings(self):
        msg1 = "2012-11-06 18:30:10 [13887] [INFO] Listening at: "
        msg2 = "http://127.0.0.1:8000 "
        msg3 = "(13887)\n"
        self.assertEqual([], self.stream._get_messages(msg1))
        self.assertEqual([], self.stream._get_messages(msg2))
        self.assertEqual(["Listening at: http://127.0.0.1:8000 (13887)\n"],
                         self.stream._get_messages(msg3))

    def test_default_max_buffer_size(self):
        self.assertEqual(10240, self.stream._max_buffer_size)

    @mock.patch("tsuru_unit_agent.stream.TsuruLogWriter")
    def test_max_buffer_size_is_configurable(self, TsuruLogWriter):
        stream = Stream(max_buffer_size=500)
        self.assertEqual(500, stream._max_buffer_size)



import unittest
import mock

from requests.exceptions import ConnectionError
from tsuru_unit_agent.main import parse_args, main


class TestMain(unittest.TestCase):
    def test_parse_args(self):
        args = parse_args(['a', 'b', 'c', 'd', 'run'])
        self.assertEqual(args.action, 'run')
        self.assertEqual(args.url, 'a')
        self.assertEqual(args.token, 'b')
        self.assertEqual(args.app_name, 'c')
        self.assertEqual(args.start_cmd, 'd')

    def test_parse_args_default_action(self):
        args = parse_args(['a', 'b', 'c', 'd'])
        self.assertEqual(args.action, 'run')
        self.assertEqual(args.url, 'a')
        self.assertEqual(args.token, 'b')
        self.assertEqual(args.app_name, 'c')
        self.assertEqual(args.start_cmd, 'd')

    def test_parse_args_deploy(self):
        args = parse_args(['a', 'b', 'c', 'd', 'deploy'])
        self.assertEqual(args.action, 'deploy')
        self.assertEqual(args.url, 'a')
        self.assertEqual(args.token, 'b')
        self.assertEqual(args.app_name, 'c')
        self.assertEqual(args.start_cmd, 'd')

    def test_parse_args_invalid(self):
        self.assertRaises(SystemExit, parse_args, [])
        self.assertRaises(SystemExit, parse_args, ['a', 'b', 'c', 'd', 'e'])

    @mock.patch('sys.argv', ['', 'http://localhost', 'token', 'app1', 'mycmd', 'deploy'])
    @mock.patch('tsuru_unit_agent.main.tasks')
    @mock.patch('tsuru_unit_agent.main.Client')
    def test_main_deploy_action(self, client_mock, tasks_mock):
        register_mock = client_mock.return_value.register_unit
        register_mock.return_value = {'env1': 'val1'}
        exec_script_mock = tasks_mock.execute_start_script
        load_yaml_mock = tasks_mock.load_app_yaml
        load_yaml_mock.return_value = {'hooks': {'build': ['cmd_1', 'cmd_2']}}
        post_app_yaml_mock = client_mock.return_value.post_app_yaml
        load_procfile_mock = tasks_mock.load_procfile
        load_procfile_mock.return_value = 'web: python myproject.py\nworker: ./startworker'
        run_build_hooks_mock = tasks_mock.run_build_hooks
        write_circus_conf_mock = tasks_mock.write_circus_conf
        save_apprc_mock = tasks_mock.save_apprc_file
        main()
        call_count = len(client_mock.mock_calls) + len(tasks_mock.mock_calls)
        self.assertEqual(call_count, 10)
        client_mock.assert_called_once_with('http://localhost', 'token')
        register_mock.assert_any_call('app1')
        v = load_yaml_mock.return_value
        v['procfile'] = load_procfile_mock.return_value
        register_mock.assert_any_call('app1', v)
        save_apprc_mock.assert_called_once_with(register_mock.return_value)
        exec_script_mock.assert_called_once_with('mycmd')
        load_yaml_mock.assert_called_once_with()
        write_circus_conf_mock.assert_called_once_with(envs={'env1': 'val1'})
        post_app_yaml_mock.assert_called_once_with('app1', load_yaml_mock.return_value)
        run_build_hooks_mock.assert_called_once_with(load_yaml_mock.return_value,
                                                     envs={'env1': 'val1'})

    @mock.patch('sys.argv', ['', 'http://localhost', 'token', 'app1', 'mycmd', 'run'])
    @mock.patch('tsuru_unit_agent.main.tasks')
    @mock.patch('tsuru_unit_agent.main.Client')
    def test_main_run_action(self, client_mock, tasks_mock):
        register_mock = client_mock.return_value.register_unit
        register_mock.return_value = {'env1': 'val1'}
        save_apprc_mock = tasks_mock.save_apprc_file
        exec_script_mock = tasks_mock.execute_start_script
        run_restart_hooks_mock = tasks_mock.run_restart_hooks
        write_circus_conf_mock = tasks_mock.write_circus_conf
        load_yaml_mock = tasks_mock.load_app_yaml
        load_yaml_mock.return_value = {'hooks': {'build': ['cmd_1', 'cmd_2']}}
        main()
        call_count = len(client_mock.mock_calls) + len(tasks_mock.mock_calls)
        self.assertEqual(call_count, 8)
        write_circus_conf_mock.assert_called_once_with(envs={'env1': 'val1'})
        client_mock.assert_called_once_with('http://localhost', 'token')
        register_mock.assert_called_once_with('app1')
        save_apprc_mock.assert_called_once_with(register_mock.return_value)
        exec_script_mock.assert_called_once_with('mycmd', envs={'env1': 'val1'}, with_shell=False)
        load_yaml_mock.assert_called_once_with()
        run_restart_hooks_mock.assert_any_call('before', load_yaml_mock.return_value,
                                               envs={'env1': 'val1'})
        run_restart_hooks_mock.assert_any_call('after', load_yaml_mock.return_value,
                                               envs={'env1': 'val1'})

    @mock.patch('sys.argv', ['', 'http://localhost', 'token', 'app1', 'mycmd', 'run'])
    @mock.patch('tsuru_unit_agent.main.tasks')
    @mock.patch('tsuru_unit_agent.main.Client')
    def test_main_run_action_api_error(self, client_mock, tasks_mock):
        register_mock = client_mock.return_value.register_unit

        def fail(*args):
            raise ConnectionError()
        register_mock.side_effect = fail
        save_apprc_mock = tasks_mock.save_apprc_file
        parse_apprc_mock = tasks_mock.parse_apprc_file
        parse_apprc_mock.return_value = {'env1': 'val1'}
        exec_script_mock = tasks_mock.execute_start_script
        run_restart_hooks_mock = tasks_mock.run_restart_hooks
        write_circus_conf_mock = tasks_mock.write_circus_conf
        load_yaml_mock = tasks_mock.load_app_yaml
        load_yaml_mock.return_value = {'hooks': {'build': ['cmd_1', 'cmd_2']}}
        main()
        call_count = len(client_mock.mock_calls) + len(tasks_mock.mock_calls)
        self.assertEqual(call_count, 8)
        write_circus_conf_mock.assert_called_once_with(envs={'env1': 'val1'})
        client_mock.assert_called_once_with('http://localhost', 'token')
        register_mock.assert_called_once_with('app1')
        parse_apprc_mock.assert_called_once_with()
        self.assertEqual(save_apprc_mock.call_count, 0)
        exec_script_mock.assert_called_once_with('mycmd', envs={'env1': 'val1'}, with_shell=False)
        load_yaml_mock.assert_called_once_with()
        run_restart_hooks_mock.assert_any_call('before', load_yaml_mock.return_value,
                                               envs={'env1': 'val1'})
        run_restart_hooks_mock.assert_any_call('after', load_yaml_mock.return_value,
                                               envs={'env1': 'val1'})



# Copyright 2014 tsuru-unit-agent authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import logging
import unittest

from logging import handlers
from socket import AF_INET, SOCK_STREAM, error

import mock

from tsuru_unit_agent import syslog


class SysLogHandlerTestCase(unittest.TestCase):

    def test_inherits_from_stdlib_syslog(self):
        assert issubclass(syslog.SysLogHandler, handlers.SysLogHandler)

    @mock.patch("socket.socket")
    def test_emit_reconnects_in_case_of_failure(self, socket_mock):
        socket_mock.return_value = reconnected_socket = mock.Mock()
        handler = syslog.SysLogHandler(address=("127.0.0.1", 3030), facility="local0",
                                       socktype=SOCK_STREAM)
        handler.socket = socket = mock.Mock()
        socket.sendall.side_effect = error("something went wrong")
        record = logging.LogRecord("name", "info", "/file.py", 15, "wow", "", "")
        handler.emit(record)
        socket_mock.assert_called_with(AF_INET, SOCK_STREAM)
        reconnected_socket.connect.assert_called_with(("127.0.0.1", 3030))
        reconnected_socket.sendall.assert_called_with(mock.ANY)



# Copyright 2015 tsuru-unit-agent authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

from setuptools import setup, find_packages
from tsuru_unit_agent import __version__

setup(
    name="tsuru-unit-agent",
    url="https://github.com/tsuru/tsuru-unit-agent",
    version=__version__,
    packages=find_packages(),
    description="Tsuru unit agent.",
    author="tsuru",
    author_email="tsuru@corp.globo.com",
    include_package_data=True,
    install_requires=["requests", "PyYAML", "honcho==0.5.0"],
    entry_points={
        'console_scripts': [
            'tsuru_unit_agent = tsuru_unit_agent.main:main',
        ],
    },
)



from socket import gethostname
import json

import requests


class Client(object):
    def __init__(self, url, token):
        self.url = url
        self.token = token

    def register_unit(self, app, custom_data=None):
        params = {
            'headers': {"Authorization": "bearer {}".format(self.token)},
        }
        request_data = {
            "hostname": gethostname(),
        }
        if custom_data is not None:
            request_data["customdata"] = json.dumps(custom_data)
        response = requests.post(
            "{}/apps/{}/units/register".format(self.url, app),
            data=request_data,
            **params)
        if 400 <= response.status_code < 500:
            response = requests.get(
                "{}/apps/{}/env".format(self.url, app),
                **params)
        if not 200 <= response.status_code < 400:
            raise Exception("invalid response {} - {}".format(response.status_code, response.text))
        tsuru_envs = response.json()
        envs = {env['name']: env['value'] for env in tsuru_envs}
        # TODO(fss): tsuru should handle this, see
        # https://github.com/tsuru/tsuru/issues/995.
        envs["port"] = envs["PORT"] = "8888"
        return envs

    def post_app_yaml(self, app, data):
        requests.post(
            "{}/apps/{}/customdata".format(self.url, app),
            data=json.dumps(data),
            headers={
                "Authorization": "bearer {}".format(self.token),
                "Content-Type": "application/json",
            },
        )



# Copyright 2015 tsuru-unit-agent authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import json
import logging
import Queue
import re
import socket
import os
import threading

from socket import gethostname

import requests

from . import syslog

QUEUE_DONE_MESSAGE = object()


def extract_message(msg):
    # 2012-11-06 18:30:10 [13887] [INFO]
    regex = "\d+\-\d+\-\d+ \d+\:\d+\:\d+ \[\d+\] \[\w+\] "
    msgs = re.split(regex, msg)
    return [m for m in msgs if m]


class Stream(object):

    def __init__(self, **kwargs):
        self._buffer = ""
        self._max_buffer_size = kwargs.get("max_buffer_size", 10240)
        self.watcher_name = kwargs.get("watcher_name", "")
        self.timeout = kwargs.get("timeout", 2)
        self.echo = kwargs.get("echo_output")
        self.default_stream_name = kwargs.get("default_stream_name", "stdout")
        envs = kwargs.get("envs") or {}
        self.envs = {}
        self.envs.update(os.environ)
        self.envs.update(envs)
        self.hostname = gethostname()
        self.start_writer()

    def start_writer(self):
        _, _, token, _, _, _, _ = self._load_envs()
        session = requests.Session()
        if token:
            session.headers.update({"Authorization": "bearer " + token})
        maxsize = int(self.envs.get("LOG_MAX_QUEUE_SIZE", 1000))
        self.queue = Queue.Queue(maxsize=maxsize)
        self.writer = TsuruLogWriter(session, self.queue)
        self.writer.start()

    def write(self, message):
        self({'data': message})
        if self.echo:
            self.echo.write(message)

    def flush(self):
        self._flush()
        if self.echo:
            self.echo.flush()

    def close(self):
        self.queue.put_nowait(QUEUE_DONE_MESSAGE)

    def __call__(self, data):
        (appname, host, token, syslog_server, syslog_port,
         syslog_facility, syslog_socket) = self._load_envs()
        messages = self._get_messages(data["data"])
        stream_name = data.get('name', self.default_stream_name)
        if appname and host and token:
            self._log_tsuru_api(messages, appname, host, token)
        if syslog_server and syslog_port and syslog_facility:
            self._log_syslog(messages, appname, syslog_server, syslog_port,
                             syslog_facility, syslog_socket, stream_name)

    def _log_tsuru_api(self, messages, appname, host, token):
        url = "{0}/apps/{1}/log?source={2}&unit={3}".format(host, appname,
                                                            self.watcher_name,
                                                            self.hostname)
        self.queue.put_nowait(LogEntry(url, self.timeout, messages))

    def _get_syslog(self, host, port, facility, socktype):
        if not hasattr(self, "_syslog"):
            if socktype == 'tcp':
                socket_type = socket.SOCK_STREAM
            else:
                socket_type = socket.SOCK_DGRAM
            self._syslog = syslog.SysLogHandler(address=(host, int(port)),
                                                facility=facility,
                                                socktype=socket_type)
            formatter = logging.Formatter('%(asctime)s {0} %(name)s: %(message)s'.format(self.hostname),
                                          "%b %d %H:%M:%S")
            self._syslog.setFormatter(formatter)
        return self._syslog

    def _log_syslog(self, messages, appname, host, port, facility, socktype, stream_name):
        try:
            syslog = self._get_syslog(host, port, facility, socktype)
            logger = logging.getLogger(appname)
            logger.handlers = []
            logger.setLevel(logging.INFO)
            logger.addHandler(syslog)
            for m in messages:
                if stream_name == 'stdout':
                    logger.info(m)
                else:
                    logger.error(m)
        except:
            pass

    def _load_envs(self):
        return (self.envs.get("TSURU_APPNAME"), self.envs.get("TSURU_HOST"),
                self.envs.get("TSURU_APP_TOKEN"), self.envs.get("TSURU_SYSLOG_SERVER"),
                self.envs.get("TSURU_SYSLOG_PORT"),
                self.envs.get("TSURU_SYSLOG_FACILITY"),
                self.envs.get("TSURU_SYSLOG_SOCKET"))

    def _flush(self):
        if len(self._buffer) > 0:
            self._buffer += "\n"
        self.write("")

    def _get_messages(self, msg):
        result = []
        if self._buffer != "":
            msg = self._buffer + msg
            self._buffer = ""
        msgs = extract_message(msg)
        lines = "".join(msgs).splitlines(True)
        for line in lines:
            if (line.endswith("\n") or len(line) > self._max_buffer_size):
                result.append(line)
            else:
                self._buffer = line
        return result


class TsuruLogWriter(threading.Thread):

    def __init__(self, session, queue, *args, **kwargs):
        super(TsuruLogWriter, self).__init__(*args, **kwargs)
        self.queue = queue
        self.session = session

    def run(self):
        while True:
            try:
                entry = self.queue.get()
                if entry == QUEUE_DONE_MESSAGE:
                    break
                try:
                    self.session.post(entry.url, data=json.dumps(entry.messages),
                                      timeout=entry.timeout)
                finally:
                    self.queue.task_done()
            except:
                pass


class LogEntry(object):

    def __init__(self, url, timeout, messages):
        self.url = url
        self.timeout = timeout
        self.messages = messages
